@@ -1,8 +1,23 @@
-import sys, string, time, copy, gc
+import sys, re, string, time, copy, gc
 from itertools import *
-from StringIO import StringIO
 import time
 
+try:
+    from io import BytesIO
+except ImportError:
+    from StringIO import StringIO as BytesIO
+
+try:
+    izip
+except NameError:
+    izip = zip  # Py3
+
+def exec_(code, glob):
+    if sys.version_info[0] >= 3:
+        exec(code, glob)
+    else:
+        exec("exec code in glob")
+
 
 TREE_FACTOR = 1 # increase tree size with '-l / '-L' cmd option
 
@@ -18,6 +33,7 @@ _ATTRIBUTES = {
 
 
 def initArgs(argv):
+    global TREE_FACTOR
     try:
         argv.remove('-l')
         # use large trees
@@ -78,6 +94,16 @@ def serialized(function):
     function.STRING = True
     return function
 
+def children(function):
+    "Decorator for benchmarks that require a list of root children"
+    function.CHILDREN = True
+    return function
+
+def nochange(function):
+    "Decorator for benchmarks that do not change the XML tree"
+    function.NO_CHANGE = True
+    return function
+
 ############################################################
 # benchmark baseclass
 ############################################################
@@ -85,8 +111,11 @@ def serialized(function):
 class SkippedTest(Exception):
     pass
 
-class BenchMarkBase(object):
+class TreeBenchMark(object):
     atoz = string.ascii_lowercase
+    repeat100  = range(100)
+    repeat500  = range(500)
+    repeat1000 = range(1000)
 
     _LIB_NAME_MAP = {
         'etree'        : 'lxe',
@@ -105,16 +134,21 @@ class BenchMarkBase(object):
             deepcopy = copy.deepcopy
             def set_property(root, fname):
                 xml = self._serialize_tree(root)
-                setattr(self, fname, lambda : etree.XML(xml, etree_parser))
+                if etree_parser is not None:
+                    setattr(self, fname, lambda : etree.XML(xml, etree_parser))
+                else:
+                    setattr(self, fname, lambda : deepcopy(root))
                 setattr(self, fname + '_xml', lambda : xml)
+                setattr(self, fname + '_children', lambda : root[:])
         else:
             def set_property(root, fname):
                 setattr(self, fname, self.et_make_clone_factory(root))
                 xml = self._serialize_tree(root)
                 setattr(self, fname + '_xml', lambda : xml)
+                setattr(self, fname + '_children', lambda : root[:])
 
-        attribute_list = list(izip(count(), ({}, _ATTRIBUTES)))
-        text_list = list(izip(count(), (None, _TEXT, _UTEXT)))
+        attribute_list = list(enumerate( [{}, _ATTRIBUTES] ))
+        text_list = list(enumerate( [None, _TEXT, _UTEXT] ))
         build_name = self._tree_builder_name
 
         self.setup_times = []
@@ -131,14 +165,16 @@ class BenchMarkBase(object):
     def _tree_builder_name(self, tree, tn, an):
         return '_root%d_T%d_A%d' % (tree, tn, an)
 
-    def tree_builder(self, tree, tn, an, serial):
+    def tree_builder(self, tree, tn, an, serial, children):
         name = self._tree_builder_name(tree, tn, an)
         if serial:
             name += '_xml'
+        elif children:
+            name += '_children'
         return getattr(self, name)
 
     def _serialize_tree(self, root):
-        return self.etree.tostring(root, 'UTF-8')
+        return self.etree.tostring(root, encoding='UTF-8')
 
     def et_make_clone_factory(self, elem):
         def generate_elem(append, elem, level):
@@ -165,8 +201,9 @@ class BenchMarkBase(object):
             "Element"    : self.etree.Element,
             "SubElement" : self.etree.SubElement
             }
+
         # create function object
-        exec "\n".join(output) in namespace
+        exec_("\n".join(output), namespace)
         return namespace["element_factory"]
 
     def _all_trees(self):
@@ -187,8 +224,9 @@ class BenchMarkBase(object):
             el = SubElement(root, "{abc}"+ch1*5, attributes)
             el.text = text
             for ch2 in atoz:
+                tag = "{cdefg}%s00001" % ch2
                 for i in range(20 * TREE_FACTOR):
-                    SubElement(el, "{cdefg}%s%05d" % (ch2, i))
+                    SubElement(el, tag).tail = text
         t = current_time() - t
         return (root, t)
 
@@ -204,7 +242,7 @@ class BenchMarkBase(object):
                 el = SubElement(root, "{abc}"+ch1*5, attributes)
                 el.text = text
                 for ch2 in atoz:
-                    SubElement(el, "{cdefg}%s%05d" % (ch2, i))
+                    SubElement(el, "{cdefg}%s00001" % ch2).tail = text
         t = current_time() - t
         return (root, t)
 
@@ -216,11 +254,11 @@ class BenchMarkBase(object):
         root = self.etree.Element('{abc}rootnode')
         children = [root]
         for i in range(6 + TREE_FACTOR):
-            tag_no = count().next
-            children = [ SubElement(c, "{cdefg}a%05d" % i, attributes)
+            children = [ SubElement(c, "{cdefg}a%05d" % (i%8), attributes)
                          for i,c in enumerate(chain(children, children, children)) ]
-        for child in root:
+        for child in children:
             child.text = text
+            child.tail = text
         t = current_time() - t
         return (root, t)
 
@@ -230,12 +268,11 @@ class BenchMarkBase(object):
         current_time = time.time
         t = current_time()
         root = self.etree.Element('{abc}rootnode')
-        children = [root]
         for ch1 in self.atoz:
             el = SubElement(root, "{abc}"+ch1*5, attributes)
             el.text = text
-            SubElement(el, "{cdefg}a00001", attributes)
-            SubElement(el, "{cdefg}z00000", attributes)
+            SubElement(el, "{cdefg}a00001", attributes).tail = text
+            SubElement(el, "{cdefg}z00000", attributes).tail = text
         t = current_time() - t
         return (root, t)
 
@@ -260,23 +297,28 @@ class BenchMarkBase(object):
             else:
                 tree_sets = ()
             if tree_sets:
-                tree_tuples = [ map(int, tree_set.split(','))
-                                for tree_set in tree_sets ]
+                tree_tuples = [list(map(int, tree_set.split(',')))
+                               for tree_set in tree_sets]
             else:
                 try:
-                    function = getattr(method, 'im_func', method)
                     arg_count = method.func_code.co_argcount - 1
                 except AttributeError:
-                    arg_count = 1
+                    try:
+                        arg_count = method.__code__.co_argcount - 1
+                    except AttributeError:
+                        arg_count = 1
                 tree_tuples = self._permutations(all_trees, arg_count)
 
-            serialized = getattr(method, 'STRING', False)
+            serialized = getattr(method, 'STRING',   False)
+            children   = getattr(method, 'CHILDREN', False)
+            no_change  = getattr(method, 'NO_CHANGE', False)
 
             for tree_tuple in tree_tuples:
                 for tn in sorted(getattr(method, 'TEXT', (0,))):
                     for an in sorted(getattr(method, 'ATTRIBUTES', (0,))):
                         benchmarks.append((name, method_call, tree_tuple,
-                                           tn, an, serialized))
+                                           tn, an, serialized, children,
+                                           no_change))
 
         return benchmarks
 
@@ -300,90 +342,107 @@ class BenchMarkBase(object):
 ############################################################
 
 def buildSuites(benchmark_class, etrees, selected):
-    benchmark_suites = map(benchmark_class, etrees)
+    benchmark_suites = list(map(benchmark_class, etrees))
 
     # sorted by name and tree tuple
     benchmarks = [ sorted(b.benchmarks()) for b in benchmark_suites ]
 
+    selected = [ re.compile(r).search for r in selected ]
+
     if selected:
         benchmarks = [ [ b for b in bs
                          if [ match for match in selected
-                              if match in b[0] ] ]
+                              if match(b[0]) ] ]
                        for bs in benchmarks ]
 
     return (benchmark_suites, benchmarks)
 
-def build_treeset_name(trees, tn, an, serialized):
+def build_treeset_name(trees, tn, an, serialized, children):
     text = {0:'-', 1:'S', 2:'U'}[tn]
     attr = {0:'-', 1:'A'}[an]
     ser  = {True:'X', False:'T'}[serialized]
-    return "%s%s%s T%s" % (text, attr, ser, ',T'.join(imap(str, trees))[:6])
+    chd  = {True:'C', False:'R'}[children]
+    return "%s%s%s%s T%s" % (text, attr, ser, chd, ',T'.join(map(str, trees))[:6])
 
 def printSetupTimes(benchmark_suites):
-    print "Setup times for trees in seconds:"
+    print("Setup times for trees in seconds:")
     for b in benchmark_suites:
-        print "%-3s:    " % b.lib_name,
+        sys.stdout.write("%-3s:     " % b.lib_name)
         for an in (0,1):
             for tn in (0,1,2):
-                print '  %s  ' % build_treeset_name((), tn, an, False)[:2],
-        print
+                sys.stdout.write('  %s   ' %
+                    build_treeset_name((), tn, an, False, False)[:2])
+        print('')
         for i, tree_times in enumerate(b.setup_times):
-            print "     T%d:" % (i+1), ' '.join("%6.4f" % t for t in tree_times)
-    print
+            print("     T%d: %s" % (i+1, ' '.join("%6.4f" % t for t in tree_times)))
+    print('')
 
-def runBench(suite, method_name, method_call, tree_set, tn, an, serial):
+def runBench(suite, method_name, method_call, tree_set, tn, an,
+             serial, children, no_change):
     if method_call is None:
         raise SkippedTest
 
     current_time = time.time
     call_repeat = range(10)
 
-    tree_builders = [ suite.tree_builder(tree, tn, an, serial)
+    tree_builders = [ suite.tree_builder(tree, tn, an, serial, children)
                       for tree in tree_set ]
 
+    rebuild_trees = not no_change and not serial
+
+    args = tuple([ build() for build in tree_builders ])
+    method_call(*args) # run once to skip setup overhead
+
     times = []
-    args = ()
     for i in range(3):
         gc.collect()
         gc.disable()
-        t = 0
+        t = -1
         for i in call_repeat:
-            args = [ build() for build in tree_builders ]
+            if rebuild_trees:
+                args = [ build() for build in tree_builders ]
             t_one_call = current_time()
             method_call(*args)
-            t += current_time() - t_one_call
-        t = 1000.0 * t / len(call_repeat)
-        times.append(t)
+            t_one_call = current_time() - t_one_call
+            if t < 0:
+                t = t_one_call
+            else:
+                t = min(t, t_one_call)
+        times.append(1000.0 * t)
         gc.enable()
-        del args
+        if rebuild_trees:
+            args = ()
+    args = ()
+    gc.collect()
     return times
 
+
 def runBenchmarks(benchmark_suites, benchmarks):
     for bench_calls in izip(*benchmarks):
         for lib, (bench, benchmark_setup) in enumerate(izip(benchmark_suites, bench_calls)):
             bench_name = benchmark_setup[0]
-            tree_set_name = build_treeset_name(*benchmark_setup[-4:])
-            print "%-3s: %-28s" % (bench.lib_name, bench_name[6:34]),
-            print "(%-10s)" % tree_set_name,
+            tree_set_name = build_treeset_name(*benchmark_setup[-6:-1])
+            sys.stdout.write("%-3s: %-28s (%-10s) " % (
+                bench.lib_name, bench_name[6:34], tree_set_name))
             sys.stdout.flush()
 
             try:
                 result = runBench(bench, *benchmark_setup)
             except SkippedTest:
-                print "skipped"
+                print("skipped")
             except KeyboardInterrupt:
-                print "interrupted by user"
+                print("interrupted by user")
                 sys.exit(1)
-            except Exception, e:
-                print "failed: %s: %s" % (e.__class__.__name__, e)
+            except Exception:
+                exc_type, exc_value = sys.exc_info()[:2]
+                print("failed: %s: %s" % (exc_type.__name__, exc_value))
+                exc_type = exc_value = None
             else:
-                print "%9.4f msec/pass, best of (" % min(result),
-                for t in result:
-                    print "%9.4f" % t,
-                print ")"
+                print("%9.4f msec/pass, best of (%s)" % (
+                      min(result), ' '.join("%9.4f" % t for t in result)))
 
         if len(benchmark_suites) > 1:
-            print # empty line between different benchmarks
+            print('')  # empty line between different benchmarks
 
 ############################################################
 # Main program
@@ -427,9 +486,8 @@ def main(benchmark_class):
             pass
         else:
             # use fast element creation in lxml.etree
-            from lxml.elements import classlookup
-            classlookup.setElementClassLookup(
-                classlookup.ElementDefaultClassLookup())
+            etree.set_element_class_lookup(
+                etree.ElementDefaultClassLookup())
 
     if len(sys.argv) > 1:
         if '-a' in sys.argv or '-c' in sys.argv:
@@ -439,11 +497,11 @@ def main(benchmark_class):
             except ValueError:
                 pass
             try:
-                import xml.etree.cElementTree as cET
+                import cElementTree as cET
                 _etrees.append(cET)
             except ImportError:
                 try:
-                    import cElementTree as cET
+                    import xml.etree.cElementTree as cET
                     _etrees.append(cET)
                 except ImportError:
                     pass
@@ -455,32 +513,33 @@ def main(benchmark_class):
             pass
         else:
             try:
-                from xml.etree import ElementTree as ET
+                from elementtree import ElementTree as ET
                 _etrees.append(ET)
             except ImportError:
                 try:
-                    from elementtree import ElementTree as ET
+                    from xml.etree import ElementTree as ET
                     _etrees.append(ET)
                 except ImportError:
                     pass
 
     if not _etrees:
-        print "No library to test. Exiting."
+        print("No library to test. Exiting.")
         sys.exit(1)
 
-    print "Preparing test suites and trees ..."
+    print("Preparing test suites and trees ...")
     selected = set( sys.argv[1:] )
     benchmark_suites, benchmarks = \
                       buildSuites(benchmark_class, _etrees, selected)
 
-    print "Running benchmark on", ', '.join(b.lib_name
-                                            for b in benchmark_suites)
-    print
+    print("Running benchmark on", ', '.join(b.lib_name
+                                            for b in benchmark_suites))
+    print('')
 
     printSetupTimes(benchmark_suites)
 
     if callgrind_zero:
         cmd = open("callgrind.cmd", 'w')
+        cmd.write('+Instrumentation\n')
         cmd.write('Zero\n')
         cmd.close()
 
